@@ -8,15 +8,15 @@
     :license: BSD, see LICENSE for details.
 
 """
-
-import __builtin__
-
+import six
+from six.moves import builtins
+from six.moves import reduce
 
 def import_object(import_name):
     module_name, expr = import_name.split(':', 1)
     mod = __import__(module_name)
     mod = reduce(getattr, module_name.split('.')[1:], mod)
-    globals = __builtin__
+    globals = builtins
     if not isinstance(globals, dict):
         globals = globals.__dict__
     return eval(expr, globals, mod.__dict__)
@@ -24,7 +24,7 @@ def import_object(import_name):
 
 def http_directive(method, path, content):
     method = method.lower().strip()
-    if isinstance(content, basestring):
+    if isinstance(content, six.string_types):
         content = content.splitlines()
     yield ''
     yield '.. http:{method}:: {path}'.format(**locals())
